..
    Note: this file is actually not auto-generated. See
    https://bugs.launchpad.net/manila/+bug/1713062.

.. _manila-compute:

.. list-table:: Description of Compute configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[nova]**
     -
   * - ``api_microversion`` = ``2.10``
     - (String) Version of Nova API to be used.
   * - ``endpoint_type`` = ``publicURL``
     - (String) Endpoint type to be used with nova client calls.
   * - ``region_name`` = ``None``
     - (String) Region name for connecting to nova.
