..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _manila-emc:

.. list-table:: Description of EMC share driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``emc_nas_login`` = ``None``
     - (String) User name for the EMC server.
   * - ``emc_nas_password`` = ``None``
     - (String) Password for the EMC server.
   * - ``emc_nas_root_dir`` = ``None``
     - (String) The root directory where shares will be located.
   * - ``emc_nas_server`` = ``None``
     - (String) EMC server hostname or IP address.
   * - ``emc_nas_server_container`` = ``None``
     - (String) DEPRECATED: Storage processor to host the NAS server. Obsolete. Unity driver supports nas server auto load balance.
   * - ``emc_nas_server_port`` = ``8080``
     - (Port number) Port number for the EMC server.
   * - ``emc_nas_server_secure`` = ``True``
     - (Boolean) Use secure connection to server.
   * - ``emc_share_backend`` = ``None``
     - (String) Share backend.
