..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _manila-quota:

.. list-table:: Description of Quota configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``max_age`` = ``0``
     - (Integer) Number of seconds between subsequent usage refreshes.
   * - ``max_gigabytes`` = ``10000``
     - (Integer) Maximum number of volume gigabytes to allow per host.
   * - ``quota_driver`` = ``manila.quota.DbQuotaDriver``
     - (String) Default driver to use for quota checks.
   * - ``quota_gigabytes`` = ``1000``
     - (Integer) Number of share gigabytes allowed per project.
   * - ``quota_share_networks`` = ``10``
     - (Integer) Number of share-networks allowed per project.
   * - ``quota_shares`` = ``50``
     - (Integer) Number of shares allowed per project.
   * - ``quota_snapshot_gigabytes`` = ``1000``
     - (Integer) Number of snapshot gigabytes allowed per project.
   * - ``quota_snapshots`` = ``50``
     - (Integer) Number of share snapshots allowed per project.
   * - ``quota_share_groups`` = ``50``
     - (Integer) Number of share groups allowed.
   * - ``quota_share_group_snapshots`` = ``50``
     - (Integer) Number of share group snapshots allowed.
   * - ``reservation_expire`` = ``86400``
     - (Integer) Number of seconds until a reservation expires.
