..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _manila-ganesha:

.. list-table:: Description of Ganesha configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``ganesha_config_dir`` = ``/etc/ganesha``
     - (String) Directory where Ganesha config files are stored.
   * - ``ganesha_config_path`` = ``$ganesha_config_dir/ganesha.conf``
     - (String) Path to main Ganesha config file.
   * - ``ganesha_db_path`` = ``$state_path/manila-ganesha.db``
     - (String) Location of Ganesha database file. (Ganesha module only.)
   * - ``ganesha_export_dir`` = ``$ganesha_config_dir/export.d``
     - (String) Path to directory containing Ganesha export configuration. (Ganesha module only.)
   * - ``ganesha_export_template_dir`` = ``/etc/manila/ganesha-export-templ.d``
     - (String) Path to directory containing Ganesha export block templates. (Ganesha module only.)
   * - ``ganesha_service_name`` = ``ganesha.nfsd``
     - (String) Name of the ganesha nfs service.
