.. -*- rst -*-

=================
Experimental APIs
=================

.. important::

    The following APIs are part of the `experimental feature <https://docs.
    openstack.org/manila/latest/contributor/experimental_apis.html>`_ introduced
    in version 2.4. The APIs may change or be removed in future versions of the
    Shared File Systems API. All experimental APIs require the
    ``X-OpenStack-Manila-API-Experimental: True`` header to be sent in the
    requests.
