..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _manila-infortrend:

.. list-table:: Description of Infortrend Manila driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``infortrend_nas_ip`` = ``None``
     - (String) Infortrend NAS ip. It is the ip for management.
   * - ``infortrend_nas_user`` = ``manila``
     - (String) Infortrend NAS username.
   * - ``infortrend_nas_password`` = ``None``
     - (String) Password for the Infortrend NAS server. This is not necessary if infortrend_nas_ssh_key is set.
   * - ``infortrend_nas_ssh_key`` = ``None``
     - (String) SSH key for the Infortrend NAS server. This is not necessary if infortrend_nas_password is set.
   * - ``infortrend_share_pools`` = ``None``
     - (String) Infortrend nas pool name list. It is separated with comma.
   * - ``infortrend_share_channels`` = ``None``
     - (String) Infortrend channels for file service. It is separated with comma.
   * - ``infortrend_cli_timeout`` = ``30``
     - (Integer) CLI timeout in seconds.
