..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _manila-spectrumscale_knfs:

.. list-table:: Description of IBM Spectrum Scale KNFS share driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description

   * - **[DEFAULT]**
     -

   * - ``gpfs_mount_point_base`` = ``$state_path/mnt``

     - (String) Base folder where exported shares are located.

   * - ``gpfs_nfs_server_list`` = ``None``

     - (List) A list of the fully qualified NFS server names that make up the OpenStack Manila configuration.

   * - ``gpfs_nfs_server_type`` = ``CES``

     - (String) NFS Server type. Valid choices are "CES" (Ganesha NFS) or "KNFS" (Kernel NFS).

   * - ``gpfs_share_export_ip`` = ``None``

     - (Host address) IP to be added to GPFS export string.

   * - ``gpfs_share_helpers`` = ``KNFS=manila.share.drivers.ibm.gpfs.KNFSHelper, CES=manila.share.drivers.ibm.gpfs.CESHelper``

     - (List) Specify list of share export helpers.

   * - ``gpfs_ssh_login`` = ``None``

     - (String) GPFS server SSH login name.

   * - ``gpfs_ssh_password`` = ``None``

     - (String) GPFS server SSH login password. The password is not needed, if 'gpfs_ssh_private_key' is configured.

   * - ``gpfs_ssh_port`` = ``22``

     - (Port number) GPFS server SSH port.

   * - ``gpfs_ssh_private_key`` = ``None``

     - (String) Path to GPFS server SSH private key for login.

   * - ``is_gpfs_node`` = ``False``

     - (Boolean) True:when Manila services are running on one of the Spectrum Scale node. False:when Manila services are not running on any of the Spectrum Scale node.
