..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _manila-hdfs:

.. list-table:: Description of HDFS share driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``hdfs_namenode_ip`` = ``None``
     - (String) The IP of the HDFS namenode.
   * - ``hdfs_namenode_port`` = ``9000``
     - (Port number) The port of HDFS namenode service.
   * - ``hdfs_ssh_name`` = ``None``
     - (String) HDFS namenode ssh login name.
   * - ``hdfs_ssh_port`` = ``22``
     - (Port number) HDFS namenode SSH port.
   * - ``hdfs_ssh_private_key`` = ``None``
     - (String) Path to HDFS namenode SSH private key for login.
   * - ``hdfs_ssh_pw`` = ``None``
     - (String) HDFS namenode SSH login password, This parameter is not necessary, if 'hdfs_ssh_private_key' is configured.
