..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _manila-vnx:

.. list-table:: Description of Dell EMC VNX share driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``vnx_ethernet_ports`` = ``None``
     - (List) Comma separated list of ports that can be used for share server interfaces. Members of the list can be Unix-style glob expressions.
   * - ``vnx_server_container`` = ``None``
     - (String) Data mover to host the NAS server.
   * - ``vnx_share_data_pools`` = ``None``
     - (List) Comma separated list of pools that can be used to persist share data.
