..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _manila-glusterfs:

.. list-table:: Description of GlusterFS share driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``glusterfs_ganesha_server_ip`` = ``None``
     - (String) Remote Ganesha server node's IP address.
   * - ``glusterfs_ganesha_server_password`` = ``None``
     - (String) Remote Ganesha server node's login password. This is not required if 'glusterfs_path_to_private_key' is configured.
   * - ``glusterfs_ganesha_server_username`` = ``root``
     - (String) Remote Ganesha server node's username.
   * - ``glusterfs_mount_point_base`` = ``$state_path/mnt``
     - (String) Base directory containing mount points for Gluster volumes.
   * - ``glusterfs_nfs_server_type`` = ``Gluster``
     - (String) Type of NFS server that mediate access to the Gluster volumes (Gluster or Ganesha).
   * - ``glusterfs_path_to_private_key`` = ``None``
     - (String) Path of Manila host's private SSH key file.
   * - ``glusterfs_server_password`` = ``None``
     - (String) Remote GlusterFS server node's login password. This is not required if 'glusterfs_path_to_private_key' is configured.
   * - ``glusterfs_servers`` =
     - (List) List of GlusterFS servers that can be used to create shares. Each GlusterFS server should be of the form [remoteuser@]<volserver>, and they are assumed to belong to distinct Gluster clusters.
   * - ``glusterfs_share_layout`` = ``None``
     - (String) Specifies GlusterFS share layout, that is, the method of associating backing GlusterFS resources to shares.
   * - ``glusterfs_target`` = ``None``
     - (String) Specifies the GlusterFS volume to be mounted on the Manila host. It is of the form [remoteuser@]<volserver>:<volid>.
   * - ``glusterfs_volume_pattern`` = ``None``
     - (String) Regular expression template used to filter GlusterFS volumes for share creation. The regex template can optionally (ie. with support of the GlusterFS backend) contain the #{size} parameter which matches an integer (sequence of digits) in which case the value shall be interpreted as size of the volume in GB. Examples: "manila-share-volume-\d+$", "manila-share-volume-#{size}G-\d+$"; with matching volume names, respectively: "manila-share-volume-12", "manila-share-volume-3G-13". In latter example, the number that matches "#{size}", that is, 3, is an indication that the size of volume is 3G.
