..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _manila-winrm:

.. list-table:: Description of WinRM configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``winrm_cert_key_pem_path`` = ``~/.ssl/key.pem``
     - (String) Path to the x509 certificate key.
   * - ``winrm_cert_pem_path`` = ``~/.ssl/cert.pem``
     - (String) Path to the x509 certificate used for accessing the serviceinstance.
   * - ``winrm_conn_timeout`` = ``60``
     - (Integer) WinRM connection timeout.
   * - ``winrm_operation_timeout`` = ``60``
     - (Integer) WinRM operation timeout.
   * - ``winrm_retry_count`` = ``3``
     - (Integer) WinRM retry count.
   * - ``winrm_retry_interval`` = ``5``
     - (Integer) WinRM retry interval in seconds
   * - ``winrm_use_cert_based_auth`` = ``False``
     - (Boolean) Use x509 certificates in order to authenticate to theservice instance.
