..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _manila-zfs:

.. list-table:: Description of ZFS share driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``zfs_dataset_creation_options`` = ``None``
     - (List) Define here list of options that should be applied for each dataset creation if needed. Example: compression=gzip,dedup=off. Note that, for secondary replicas option 'readonly' will be set to 'on' and for active replicas to 'off' in any way. Also, 'quota' will be equal to share size. Optional.
   * - ``zfs_dataset_name_prefix`` = ``manila_share_``
     - (String) Prefix to be used in each dataset name. Optional.
   * - ``zfs_dataset_snapshot_name_prefix`` = ``manila_share_snapshot_``
     - (String) Prefix to be used in each dataset snapshot name. Optional.
   * - ``zfs_migration_snapshot_prefix`` = ``tmp_snapshot_for_share_migration_``
     - (String) Set snapshot prefix for usage in ZFS migration. Required.
   * - ``zfs_replica_snapshot_prefix`` = ``tmp_snapshot_for_replication_``
     - (String) Set snapshot prefix for usage in ZFS replication. Required.
   * - ``zfs_service_ip`` = ``None``
     - (String) IP to be added to admin-facing export location. Required.
   * - ``zfs_share_export_ip`` = ``None``
     - (String) IP to be added to user-facing export location. Required.
   * - ``zfs_share_helpers`` = ``NFS=manila.share.drivers.zfsonlinux.utils.NFSviaZFSHelper``
     - (List) Specify list of share export helpers for ZFS storage. It should look like following: 'FOO_protocol=foo.FooClass,BAR_protocol=bar.BarClass'. Required.
   * - ``zfs_ssh_private_key_path`` = ``None``
     - (String) Path to SSH private key that should be used for SSH'ing ZFS storage host. Not used for replication operations. Optional.
   * - ``zfs_ssh_user_password`` = ``None``
     - (String) Password for user that is used for SSH'ing ZFS storage host. Not used for replication operations. They require passwordless SSH access. Optional.
   * - ``zfs_ssh_username`` = ``None``
     - (String) SSH user that will be used in 2 cases: 1) By manila-share service in case it is located on different host than its ZFS storage. 2) By manila-share services with other ZFS backends that perform replication. It is expected that SSH'ing will be key-based, passwordless. This user should be passwordless sudoer. Optional.
   * - ``zfs_use_ssh`` = ``False``
     - (Boolean) Remote ZFS storage hostname that should be used for SSH'ing. Optional.
   * - ``zfs_zpool_list`` = ``None``
     - (List) Specify list of zpools that are allowed to be used by backend. Can contain nested datasets. Examples: Without nested dataset: 'zpool_name'. With nested dataset: 'zpool_name/nested_dataset_name'. Required.
