..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _manila-lvm:

.. list-table:: Description of LVM share driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``lvm_share_export_ips`` = ``None``
     - (String) List of IPs to export shares belonging to the LVM storage driver.
   * - ``lvm_share_export_root`` = ``$state_path/mnt``
     - (String) Base folder where exported shares are located.
   * - ``lvm_share_helpers`` = ``CIFS=manila.share.drivers.helpers.CIFSHelperUserAccess, NFS=manila.share.drivers.helpers.NFSHelper``
     - (List) Specify list of share export helpers.
   * - ``lvm_share_mirrors`` = ``0``
     - (Integer) If set, create LVMs with multiple mirrors. Note that this requires lvm_mirrors + 2 PVs with available space.
   * - ``lvm_share_volume_group`` = ``lvm-shares``
     - (String) Name for the VG that will contain exported shares.
